from spetlr.exceptions import SpetlrException


class SpetlrEhException(SpetlrException):
    pass


class SpetlrEhInitException(SpetlrEhException):
    pass


class SpetlrEhLogicException(SpetlrEhException):
    pass
